@@ -3,7 +3,7 @@
  *
  * Copyright (C) 2005-2012  NTT DATA CORPORATION
  *
- * Version: 0.1.1   2012/05/05
+ * Version: 0.1.3   2012/08/02
  */
 
 #include "internal.h"
@@ -207,6 +207,27 @@ out:
 	return in_use;
 }
 
+/**
+ * ccs_acl_info_has_sub_acl - Clear "struct ccs_acl_info"->acl_info.
+ *
+ * @list: Pointer to "struct list_head".
+ *
+ * Returns true if @list is not empty, false otherwise.
+ */
+static bool ccs_acl_info_has_sub_acl(struct list_head *list)
+{
+	struct ccs_acl_info *acl;
+	struct ccs_acl_info *tmp;
+	if (list_empty(list))
+		return false;
+	mutex_lock(&ccs_policy_lock);
+	list_for_each_entry_safe(acl, tmp, list, list) {
+		ccs_try_to_gc(CCS_ID_ACL, &acl->list);
+	}
+	mutex_unlock(&ccs_policy_lock);
+	return !list_empty(list);
+}
+
 /**
  * ccs_del_acl - Delete members in "struct ccs_acl_info".
  *
@@ -362,6 +383,14 @@ static void ccs_try_to_gc(const enum ccs_policy_id type,
 			goto reinject;
 		break;
 	case CCS_ID_ACL:
+		/*
+		 * Don't kfree() until "struct ccs_acl_info"->acl_info_list
+		 * becomes empty.
+		 */
+		if (ccs_acl_info_has_sub_acl
+		    (&container_of(element, typeof(struct ccs_acl_info),
+				   list)->acl_info_list))
+			goto reinject;
 		ccs_del_acl(element);
 		break;
 	case CCS_ID_DOMAIN:
@@ -461,8 +490,7 @@ static void ccs_collect_entry(void)
 		struct list_head * const list = &ccs_acl_list[i];
 		list_for_each_entry_safe(ptr, tmp, list, list) {
 			ccs_collect_acl(&ptr->acl_info_list);
-			if (!ptr->is_deleted ||
-			    !list_empty(&ptr->acl_info_list))
+			if (!ptr->is_deleted)
 				continue;
 			/* ptr->is_deleted = CCS_GC_IN_PROGRESS; */
 			ccs_try_to_gc(CCS_ID_ACL, &ptr->list);
